#pragma semicolon 1
#pragma newdecls required

#include <sourcemod>
#undef REQUIRE_PLUGIN
#include <hx_stats>

public Plugin myinfo = 
{
	name = "[L4D] HxStats Natives Demo",
	author = "Dragokas",
	description = "",
	version = "1.2",
	url = "https://forums.alliedmods.net/showthread.php?t=320247"
}

bool g_bHxStatsAvail;

/*
	README:
	 * following plugin demonstrates simple examples:
	  - how to retrieve the number of total points of specified target using console command "sm_getpoints"
	  - how to add points to player statistics using console command "sm_addpoints"
	  - how to print in chat the total number of tanks killed by every player when they joined, using "HX_OnClientRegistered" Forward.
*/

public void OnPluginStart()
{
	RegAdminCmd("sm_getpoints", CmdGetPoints, ADMFLAG_ROOT, "<target> Get points of specified player (or your own if none specified)");
	RegAdminCmd("sm_addpoints", CmdAddPoints, ADMFLAG_ROOT,	"<target> <points> Add points to specified player");
}

public Action CmdGetPoints(int client, int args)
{
	int target;
	
	if( args == 0 )
	{
		target = client;
	}
	else {
		char sArg[4];
		GetCmdArg(1, sArg, sizeof sArg);
		target = StringToInt(sArg);
	}

	if( g_bHxStatsAvail )
	{
		if( HX_IsClientRegistered(target) )
		{
			int points_total = HX_GetPoints(target, HX_COUNTING_ACTUAL, HX_POINTS);
			int points_round = HX_GetPoints(target, HX_COUNTING_ACTUAL, HX_POINTS);
			
			ReplyToCommand(client, "[%N] Total points: %i (round points: %i)", target, points_total, points_round);
		}
		else {
			ReplyToCommand(client, "[%N] Client: %i is not yet initialized by database. Please, wait...", target, target);
		}
	}
	else {
		ReplyToCommand(client, "HxStats is not available at the moment!");
	}
	
	return Plugin_Handled;
}

public Action CmdAddPoints(int client, int args)
{
	int target, points;
	
	if( args < 2 )
	{
		ReplyToCommand(client, "using: sm_addpoints <target> <points>");
		return Plugin_Handled;
	}
	
	char sArg[16];
	GetCmdArg(1, sArg, sizeof sArg);
	target = StringToInt(sArg);
	
	GetCmdArg(2, sArg, sizeof sArg);
	points = StringToInt(sArg);
	
	if( g_bHxStatsAvail )
	{
		if( HX_IsClientRegistered(target) )
		{
			HX_AddPoints(target, points, HX_POINTS);
			
			ReplyToCommand(client, "[%N] Added round points: %i", target, points);
		}
		else {
			ReplyToCommand(client, "[%N] Client: %i is not yet initialized by database. Please, wait...", target, target);
		}
	}
	else {
		ReplyToCommand(client, "HxStats is not available at the moment!");
	}
	
	return Plugin_Handled;
}

public void HX_OnClientRegistered(int client)
{
	if( client && IsClientInGame(client) && GetClientTeam(client) )
	{
		if( HX_IsClientRegistered(client) )
		{
			int tanks = HX_GetPoints(client, HX_COUNTING_ACTUAL, HX_TANK);
			
			PrintToChatAll("Player %N is just joined with number of tanks killed: %i", client, tanks);
		}
	}
}

public void OnAllPluginsLoaded()
{
	g_bHxStatsAvail = (GetFeatureStatus(FeatureType_Native, "HX_GetPoints") == FeatureStatus_Available);
}

public void OnLibraryAdded(const char[] name)
{
	if( strcmp(name, "hx_stats") == 0 )
	{
		g_bHxStatsAvail = true;
	}
}

public void OnLibraryRemoved(const char[] name)
{
	if( strcmp(name, "hx_stats") == 0 )
	{
		g_bHxStatsAvail = false;
	}
}